@@ -17,7 +17,7 @@ using namespace apolloron;
 
 const char *INKF_PROGNAME = "inkf";
 const char *INKF_BRIEF    = "Yet Another Network Kanji Filter";
-const char *INKF_VERSION  = "1.2.3";
+const char *INKF_VERSION  = "2.0.0";
 const char *INKF_COPY     = "Copyright (C) 2002-2018 Masashi Astro Tachibana";
 const char *INKF_DEF_OUT  = "inkf.out";
 
@@ -118,7 +118,12 @@ int main(int argc, char *argv[]) {
 
         // read from stdin
         if (option.mime_decode == MIME_NONE && option.mime_encode == MIME_NONE &&
-                !option.flag_guess && !option.flag_html_to_plain && !option.flag_midi) {
+                !option.flag_guess && !option.flag_html_to_plain && !option.flag_midi &&
+                strcasecmp(option.output_charset, "UTF-8_BOM") != 0 &&
+                strncasecmp(option.input_charset, "UTF-16", 6) != 0 &&
+                strncasecmp(option.input_charset, "UTF-32", 6) != 0 &&
+                !(!strcasecmp(option.output_charset, "UTF-16") ||
+                !strcasecmp(option.output_charset, "UTF-32"))) {
             while (!feof(fpin)) {
                 tmp_str = "";
                 while (!feof(fpin)) {
@@ -139,20 +144,21 @@ int main(int argc, char *argv[]) {
                 }
             }
         } else {
+            tmp_str = "";
             while (!feof(fpin)) {
                 l = fread(buf, 1, 4096, fpin);
                 if (l <= 0) {
                     continue;
                 }
                 buf[l] = '\0';
-                tmp_str.add(buf);
+                tmp_str.addBinary(buf, l);
             }
             if (option.flag_guess) {
                 guess(tmp_str, option.input_charset, fpout);
             } else if (option.flag_midi) {
                 tmp_str = tmp_str.toMIDI();
                 fwrite(tmp_str.c_str(), 1, tmp_str.binaryLength(), fpout);
-            } else if (0 < tmp_str.len()) {
+            } else if (0 < tmp_str.len() || 0 < tmp_str.binaryLength()) {
                 convert(tmp_str, option, fpout);
             }
         }
@@ -164,6 +170,8 @@ int main(int argc, char *argv[]) {
             charset_autodetect_pre = true;
         }
 
+        tmp_str.useAsBinary(0);
+
         // read from file(s) or specified URL
         for (i = 0; option.input_filenames[i] != NULL; i++) {
             if (charset_autodetect_pre) {
@@ -308,6 +316,26 @@ bool set_option(TOption &option, int argc, const char **argv) {
                     case 'g':
                         option.flag_guess = true;
                         break;
+                    case 'f':
+                        if (argv[i][j+1] == '\0' && i + 1 < argc) {
+                            strncpy(option.input_charset, argv[i+1], 31);
+                            option.input_charset[31] = '\0';
+                            i++;
+                            j = strlen(argv[i]) - 1;
+                        } else {
+                            return false; // invalid parameter
+                        }
+                        break;
+                    case 't':
+                        if (argv[i][j+1] == '\0' && i + 1 < argc) {
+                            strncpy(option.output_charset, argv[i+1], 31);
+                            option.output_charset[31] = '\0';
+                            i++;
+                            j = strlen(argv[i]) - 1;
+                        } else {
+                            return false; // invalid parameter
+                        }
+                        break;
                     case 'j':
                         strcpy(option.output_charset, "ISO-2022-JP");
                         break;
@@ -317,8 +345,53 @@ bool set_option(TOption &option, int argc, const char **argv) {
                     case 'e':
                         strcpy(option.output_charset, "EUC-JP");
                         break;
+                    case 'l':
+                        strcpy(option.output_charset, "LATIN1");
+                        break;
                     case 'w':
-                        strcpy(option.output_charset, "UTF-8");
+                        if (argv[i][j + 1] == '1' && argv[i][j + 2] == '6') {
+                            if (argv[i][j + 3] == 'L' && argv[i][j + 4] == '0') {
+                                strcpy(option.output_charset, "UTF-16LE");
+                                j += 4;
+                            } else if (argv[i][j + 3] == 'L') {
+                                strcpy(option.output_charset, "UTF-16LE_BOM");
+                                j += 3;
+                            } else if (argv[i][j + 3] == 'B' && argv[i][j + 4] == '0') {
+                                strcpy(option.output_charset, "UTF-16BE");
+                                j += 4;
+                            } else if (argv[i][j + 3] == 'B') {
+                                strcpy(option.output_charset, "UTF-16BE_BOM");
+                                j += 3;
+                            } else {
+                                strcpy(option.output_charset, "UTF-16BE");
+                                j += 2;
+                            }
+                        } else if (argv[i][j + 1] == '3' && argv[i][j + 2] == '2') {
+                            if (argv[i][j + 3] == 'L' && argv[i][j + 4] == '0') {
+                                strcpy(option.output_charset, "UTF-32LE");
+                                j += 4;
+                            } else if (argv[i][j + 3] == 'L') {
+                                strcpy(option.output_charset, "UTF-32LE_BOM");
+                                j += 3;
+                            } else if (argv[i][j + 3] == 'B' && argv[i][j + 4] == '0') {
+                                strcpy(option.output_charset, "UTF-32BE");
+                                j += 4;
+                            } else if (argv[i][j + 3] == 'B') {
+                                strcpy(option.output_charset, "UTF-32BE_BOM");
+                                j += 3;
+                            } else {
+                                strcpy(option.output_charset, "UTF-32BE");
+                                j += 2;
+                            }
+                        } else if (argv[i][j + 1] == '8' && argv[i][j + 2] == '0') {
+                            strcpy(option.output_charset, "UTF-8");
+                            j += 2;
+                        } else if (argv[i][j + 1] == '8') {
+                            strcpy(option.output_charset, "UTF-8_BOM");
+                            j++;
+                        } else {
+                            strcpy(option.output_charset, "UTF-8");
+                        }
                         break;
                     case 'k':
                         strcpy(option.output_charset, "EUC-KR");
@@ -326,7 +399,7 @@ bool set_option(TOption &option, int argc, const char **argv) {
                     case 'c':
                         strcpy(option.output_charset, "GB2312");
                         break;
-                    case 't':
+                    case 'b':
                         strcpy(option.output_charset, "BIG5");
                         break;
                     case 'J':
@@ -339,7 +412,34 @@ bool set_option(TOption &option, int argc, const char **argv) {
                         strcpy(option.input_charset, "EUC-JP");
                         break;
                     case 'W':
-                        strcpy(option.input_charset, "UTF-8");
+                        if (argv[i][j + 1] == '1' && argv[i][j + 2] == '6') {
+                            if (argv[i][j + 3] == 'L') {
+                                strcpy(option.input_charset, "UTF-16LE");
+                                j += 3;
+                            } else if (argv[i][j + 3] == 'B') {
+                                strcpy(option.input_charset, "UTF-16BE");
+                                j += 3;
+                            } else {
+                                strcpy(option.input_charset, "UTF-16LE");
+                                j += 2;
+                            }
+                        } else if (argv[i][j + 1] == '3' && argv[i][j + 2] == '2') {
+                            if (argv[i][j + 3] == 'L') {
+                                strcpy(option.input_charset, "UTF-32LE");
+                                j += 3;
+                            } else if (argv[i][j + 3] == 'B') {
+                                strcpy(option.input_charset, "UTF-32BE");
+                                j += 3;
+                            } else {
+                                strcpy(option.input_charset, "UTF-32LE");
+                                j += 2;
+                            }
+                        } else if (argv[i][j + 1] == '8') {
+                            strcpy(option.input_charset, "UTF-8");
+                            j++;
+                        } else {
+                            strcpy(option.input_charset, "UTF-8");
+                        }
                         break;
                     case 'K':
                         strcpy(option.input_charset, "EUC-KR");
@@ -347,7 +447,7 @@ bool set_option(TOption &option, int argc, const char **argv) {
                     case 'C':
                         strcpy(option.input_charset, "GB2312");
                         break;
-                    case 'T':
+                    case 'B':
                         strcpy(option.input_charset, "BIG5");
                         break;
                     case 'm':
@@ -395,9 +495,6 @@ bool set_option(TOption &option, int argc, const char **argv) {
                             j++;
                         }
                         break;
-                    case 'b': // ignore bufferd option
-                    case 'u': // ignore unbufferd option
-                        break;
                     default:
                         return false; // invalid parameter
                         break;
@@ -442,6 +539,8 @@ bool set_option(TOption &option, int argc, const char **argv) {
 void guess(const String &str, const char *input_charset, FILE *fpout) {
     const char *charcode;
     const char *retcode;
+    long i, length;
+    char endian = 'L';
 
     if (input_charset != NULL && !strncasecmp(input_charset, "AUTODETECT", 10)) {
         charcode = str.detectCharSet();
@@ -460,18 +559,88 @@ void guess(const String &str, const char *input_charset, FILE *fpout) {
         charcode = input_charset;
     }
 
-    retcode = "LF";
-    if (0 <= str.search(RETURN_STR[LINE_MODE_CRLF])) {
-        retcode = "CRLF";
-    } else if (0 <= str.search(RETURN_STR[LINE_MODE_CR])) {
-        retcode = "CR";
+    if ((!strncmp(charcode, "UTF-16", 6) || !strncmp(charcode, "UTF-32", 6)) &&
+            strstr(charcode, "BE") != NULL) {
+        endian = 'B';
+    } else {
+        endian = 'L';
     }
 
+    retcode = "LF";
+    if (!strncmp(charcode, "UTF-16", 6)) {
+        length = str.binaryLength();
+        for (i = 0; i < length; i += 2) {
+            if (endian == 'L' && i + 3 < length &&
+                    str[i] == '\x0D' && str[i+1] == '\x00' &&
+                    str[i+2] == '\x0A' && str[i+3] == '\x00') {
+                retcode = "CRLF";
+                break;
+            } else if (endian == 'B' && i + 3 < length &&
+                    str[i] == '\x00' && str[i+1] == '\x0D' &&
+                    str[i+2] == '\x00' && str[i+3] == '\x0A') {
+                retcode = "CRLF";
+                break;
+            }
+        }
+        if (!strcmp(retcode, "LF")) {
+            for (i = 0; i < length; i += 2) {
+                if (endian == 'L' && str[i] == '\x0D' && str[i+1] == '\x00') {
+                    retcode = "CR";
+                    break;
+                } else if (endian == 'B' && str[i] == '\x00' && str[i+1] == '\x0D') {
+                    retcode = "CR";
+                    break;
+                }
+            }
+        }
+    } else if (!strncmp(charcode, "UTF-32", 6)) {
+        length = str.binaryLength();
+        for (i = 0; i < length; i += 4) {
+            if (endian == 'L' && i + 7 < length &&
+                    str[i] == '\x0D' && str[i+1] == '\x00' &&
+                    str[i+2] == '\x00' && str[i+3] == '\x00' &&
+                    str[i+4] == '\x0A' && str[i+5] == '\x00' &&
+                    str[i+6] == '\x00' && str[i+7] == '\x00') {
+                retcode = "CRLF";
+                break;
+            } else if (endian == 'B' && i + 7 < length &&
+                    str[i] == '\x00' && str[i+1] == '\x00' &&
+                    str[i+2] == '\x00' && str[i+3] == '\x0D' &&
+                    str[i+4] == '\x00' && str[i+5] == '\x00' &&
+                    str[i+6] == '\x00' && str[i+7] == '\x0A') {
+                retcode = "CRLF";
+                break;
+            }
+        }
+        if (!strcmp(retcode, "LF")) {
+            for (i = 0; i < length; i += 4) {
+                if (endian == 'L' && i + 3 < length &&
+                        str[i] == '\x0D' && str[i+1] == '\x00' &&
+                        str[i+2] == '\x00' && str[i+3] == '\x00') {
+                    retcode = "CR";
+                    break;
+                } else if (endian == 'B' && i + 3 < length &&
+                        str[i] == '\x00' && str[i+1] == '\x00' &&
+                        str[i+2] == '\x00' && str[i+3] == '\x0D') {
+                    retcode = "CR";
+                    break;
+                }
+            }
+        }
+    } else {
+        if (0 <= str.search(RETURN_STR[LINE_MODE_CRLF])) {
+            retcode = "CRLF";
+        } else if (0 <= str.search(RETURN_STR[LINE_MODE_CR])) {
+            retcode = "CR";
+        }
+    }
     fprintf(fpout, "%s (%s)\n", charcode, retcode);
 }
 
 
 void convert(String &str, const TOption &option, FILE *fpout) {
+    String u8str;
+
     if (option.mime_decode == MIME_BASE64) {
         str = str.decodeBASE64();
     } else if (option.mime_decode == MIME_QUOTED_PRINTABLE) {
@@ -489,7 +658,10 @@ void convert(String &str, const TOption &option, FILE *fpout) {
             str = str.decodeQuotedPrintable();
         }
     }
-    str = str.strconv(option.input_charset, option.output_charset);
+
+    if (!(!strcasecmp(option.input_charset, "AUTODETECT") && option.output_charset[0] == '\0')) {
+        str = str.strconv(option.input_charset, option.output_charset);
+    }
     if (option.flag_html_to_plain) {
         str = str.convertHTMLToPlain(option.output_charset, option.output_charset);
     }
@@ -509,7 +681,9 @@ void convert(String &str, const TOption &option, FILE *fpout) {
         str = str.changeWidth("K", option.output_charset, option.output_charset);
     }
     if (option.line_mode != LINE_MODE_NOCONVERSION) {
-        str = str.changeReturnCode(RETURN_STR[option.line_mode]);
+        u8str = str.strconv(option.output_charset, "UTF-8");
+        u8str = u8str.changeReturnCode(RETURN_STR[option.line_mode]);
+        str = u8str.strconv("UTF-8", option.output_charset);
     }
     if (option.mime_encode == MIME_BASE64) {
         str = str.encodeBASE64();
@@ -523,24 +697,35 @@ void convert(String &str, const TOption &option, FILE *fpout) {
                 option.output_charset, 72, RETURN_STR[option.line_mode], 'Q');
     }
 
-    fwrite(str.c_str(), 1, str.len(), fpout);
+    if (str.isBinary()) {
+        fwrite(str.c_str(), 1, str.binaryLength(), fpout);
+    } else {
+        fwrite(str.c_str(), 1, str.len(), fpout);
+    }
 }
 
 
 void show_help() {
     fprintf(stderr, 
       "Usage:  %s -[flags] [--] [in file/in URL] .. [out file for -O flag]\n"
-      " j/s/e/w/k/c/t  Specify output encoding ISO-2022-JP, Shift_JIS, EUC-JP, UTF-8,\n"
+      " j/s/e/w/k/c/b  Specify output encoding ISO-2022-JP, Shift_JIS, EUC-JP, UTF-8,\n"
       "                EUC-KR, GB2312, BIG5\n"
-      " J/S/E/W/K/C/T  Specify input encoding ISO-2022-JP, Shift_JIS, EUC-JP, UTF-8,\n"
+      " l              LATIN1\n"
+      " w8[0]          UTF-8\n"
+      " w16/w32[BL][0] UTF-16, UTF-32\n"
+      " J/S/E/W/K/C/B  Specify input encoding ISO-2022-JP, Shift_JIS, EUC-JP, UTF-8,\n"
       "                EUC-KR, GB2312, BIG5\n"
+      " W8             UTF-8\n"
+      " W16/W32[BL]    UTF-16, UTF-32\n"
       " m[BQN0]        MIME decode [B:base64,Q:quoted,N:non-strict,0:no decode]\n"
       " M[BQ]          MIME encode [B:base64 Q:quoted]\n"
       " Z              Convert JISX0208 Alphabet to ASCII\n"
       " O              Output to File (DEFAULT '%s')\n"
       " L[uwm]         Line mode u:LF w:CRLF m:CR (DEFAULT noconversion)\n"
-      " --ic=<encoding>        Specify the input encoding\n"
-      " --oc=<encoding>        Specify the output encoding\n"
+      " t <encoding>   Specify output encoding (iconv compatible)\n"
+      " f <encoding>   Specify input encoding (iconv compatible)\n"
+      " --oc=<encoding>    Specify the output encoding\n"
+      " --ic=<encoding>    Specify the input encoding\n"
       "  *supported encoding is as follows:\n"
       "   ISO-2022-JP\n"
       "   Shift_JIS (CP932)\n"
@@ -555,6 +740,8 @@ void show_help() {
       "   UTF-8\n"
       "   UTF-7\n"
       "   UTF-7-IMAP\n"
+      "   UTF-16, UTF-16BE, UTF-16LE\n"
+      "   UTF-32, UTF-32BE, UTF-32LE\n"
       "   AUTODETECT\n"
       "   AUTODETECT_JP\n"
       " --hiragana --katakana  Hiragana/Katakana Conversion\n"
